/*=========================================================================
  PEIS.cc

  Library   : Image Registration Toolkit (IRTK)
  Copyright : Imperial College, Department of Computing
              Visual Information Processing (VIP), 2008 onwards
  Date      : $Date: 2014-04-07 20:04:03 +0100 (Mon, 07 Apr 2014) $
  Changes   : $Author: cl6311 $
=========================================================================*/

#include <irtkImage.h>
#include <irtkEuclideanDistanceTransform.h>
#include <irtkGradientImageFilter.h>
#include <irtkGaussianBlurring.h>

#define MIN_BIAS -100000
#define MAXL1 100

bool bDebug = false;
bool bNoNonLocal = false;
bool bNoSpatialWeighting = false;
int L1limit = MAXL1;

int is2D = false;
int patchSizeMM;
int patchRadiusX, patchRadiusY, patchRadiusZ;
int patchSizeX = 5;
int patchSizeY = 5;
int patchSizeZ = 5;

// similar to Coupe's SNIPE paper
double sigmaSimilarity = 0.5;
// nonlocal STAPLE
double sigmaSpatial = 2.0;
int maxNumberOfFacets = 0;
bool simTypeGaussian = false;

int nrOfL1Steps[MAXL1+1];
// the y dimension of this array is a heuristic
int xL1Steps[MAXL1+1][4*(MAXL1+1)*(MAXL1+1)];
int yL1Steps[MAXL1+1][4*(MAXL1+1)*(MAXL1+1)];
int zL1Steps[MAXL1+1][4*(MAXL1+1)*(MAXL1+1)];
int valueL1Steps[MAXL1+1][4*(MAXL1+1)*(MAXL1+1)];

irtkGreyImage labelA;
irtkGreyImage labelB;

irtkRealImage biasImg;

irtkRealImage biasImgX;
irtkRealImage biasImgY;
irtkRealImage biasImgZ;

irtkRealImage distMapB;

irtkRealImage gradBX;
irtkRealImage gradBY;
irtkRealImage gradBZ;

irtkGreyImage maskImg;

irtkRealImage variationImg;
irtkRealImage variationImgX;
irtkRealImage variationImgY;
irtkRealImage variationImgZ;

char* outputBias = NULL;

int* spatialPenalty = NULL;

void usage()
{
  cerr << "Usage: PEIS [testSeg] [referenceSeg] <options>" << endl;
  cerr << "<-patchSizeMM value> (use isotropic patch size in millimeter, by default patch size is in voxels)" << endl;
  cerr << "<-patchSizeXYZ valX valY valZ> (patch size in voxels, default 5 5 5)" << endl;
  cerr << "<-simTypeGaussian> (experimental)" << endl;
  cerr << "<-sigmaSimilarity value> (default: 0.1, if simTypeGaussian only)" << endl;
  cerr << "<-sigmaSpatial value> (default: 2.0, if simTypeGaussian only)" << endl;
  cerr << "<-noSpatialWeighting> (alpha=theta=0.5)" << endl;
  cerr << "<-noNonLocal> (do not perform a patch search)" << endl;
  cerr << "<-L1limit value> (default: 100, limit of search radius in L1 norm)" << endl;
  cerr << "<-verbose>" << endl;
  cerr << "<-biasImage output> (save bias image to output)" << endl;
  cerr << endl << "Output: PEIS (eta) PEISSimilarity (gamma) PEISSpatial (tau) mean_Bias SD mean_BiasX SD mean_BiasY SD mean_BiasZ SD" << endl;
  cerr << endl << "Version 1.0: June, 20, 2014" << endl;
  cerr << "Method described in: 'Patch-based Evaluation of Image Segmentation' by Ledig, C. et al. in CVPR 2014" << endl;
  cerr << "Author: Christian Ledig (christian.ledig@imperial.ac.uk)" << endl;
  exit(1);
}

int getVoxelsForDisplacement( int dx, int dy, int dz = 0);
double getSpatialIndex( int dx, int dy, int dz = 0 );
double getSimilarityIndex ( int incorrectLabels );
int maxNumberOfEdgeFacets( irtkRealImage& currentPatch );
void numberOfEdgeFacets( irtkRealImage& currentPatch, double* totVar, double *totVarX, double *totVarY, double *totVarZ );
void calculateDistanceTransform();
void calculateGradients();
double calculatePBOverlap();
int getIndex( int x, int y, int z);
void calculatePatchNormal( int x, int y, int z, double &xn, double &yn, double &zn );
int comparePatches( int xA, int yA, int zA, int xB ,int yB, int zB, int minValue );
void calculateL1Steps();
double calculateVectorNorm( double x, double y, double z);

int main(int argc, char **argv)
{
  if (argc < 3) {
    usage();
    exit(1);
  }

  labelA.Read( argv[1] );
  argv++;
  argc--;

  labelB.Read( argv[1] );
  argv++;
  argc--;

  while (argc > 1) {
     bool ok = false;
     if ((ok == false) && (strcmp(argv[1], "-patchSizeMM") == 0)) {
     	argc--;
     	argv++;
     	patchSizeMM = atof( argv[1] );
     	argc--;
     	argv++;
     	ok = true;

    	patchSizeX = (int) (patchSizeMM / labelB.GetXSize() + 0.5);
    	patchSizeY = (int) (patchSizeMM / labelB.GetYSize() + 0.5);
    	patchSizeZ = (int) (patchSizeMM / labelB.GetZSize() + 0.5);

    	if( patchSizeX % 2 == 0 ) patchSizeX++;
    	if( patchSizeY % 2 == 0 ) patchSizeY++;
    	if( patchSizeZ % 2 == 0 ) patchSizeZ++;
     }
     if ((ok == false) && (strcmp(argv[1], "-patchSizeXYZ") == 0)) {
     	argc--;
     	argv++;
     	patchSizeX = atoi( argv[1] );
     	argc--;
     	argv++;
     	patchSizeY = atoi( argv[1] );
     	argc--;
     	argv++;
     	patchSizeZ = atoi( argv[1] );
     	argc--;
     	argv++;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-simTypeGaussian") == 0)) {
     	argc--;
     	argv++;
     	simTypeGaussian = true;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-sigmaSimilarity") == 0)) {
     	argc--;
     	argv++;
     	sigmaSimilarity = atof( argv[1] );
     	argc--;
     	argv++;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-sigmaSpatial") == 0)) {
     	argc--;
     	argv++;
     	sigmaSpatial = atof( argv[1] );
     	argc--;
     	argv++;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-noNonLocal") == 0)) {
     	argc--;
     	argv++;
     	bNoNonLocal = true;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-noSpatialWeighting") == 0)) {
     	argc--;
     	argv++;
     	bNoSpatialWeighting = true;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-L1limit") == 0)) {
      	argc--;
      	argv++;
      	L1limit = atoi(argv[1]);
      	argc--;
      	argv++;
      	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-verbose") == 0)) {
     	argc--;
     	argv++;
     	bDebug = true;
     	ok = true;
     }
     if ((ok == false) && (strcmp(argv[1], "-biasImage") == 0)) {
         	argc--;
         	argv++;
         	outputBias = argv[1];
         	argv++;
         	argc--;
         	ok = true;
     }
     if (ok == false) {
       cerr << "Can not parse argument " << argv[1] << endl;
       usage();
     }
  }

  if( labelA.GetZ() == 1 )
  {
if( bDebug ) cerr << "2D-Mode" << endl;
	  is2D = true;
  }

	patchRadiusX = patchSizeX / 2;
	patchRadiusY = patchSizeY / 2;
	patchRadiusZ = patchSizeZ / 2;

if( bDebug ) cerr << "PatchSize XYZ: " << patchSizeX << " " << patchSizeY << " " << patchSizeZ << endl;
if( bDebug ) cerr << "PatchRadius XYZ: "  << patchRadiusX << " " << patchRadiusY << " " << patchRadiusZ << endl;

  calculateL1Steps();
  calculateGradients();
  calculatePBOverlap();

  if( spatialPenalty != NULL )
  {
	  delete[] spatialPenalty;
  }

  return 0;
}

double calculatePBOverlap()
{
	int debugMaxL1=0;

	maskImg = labelB;

	biasImg.Initialize( labelB.GetImageAttributes() );

	biasImgX.Initialize( labelB.GetImageAttributes() );
	biasImgY.Initialize( labelB.GetImageAttributes() );
	biasImgZ.Initialize( labelB.GetImageAttributes() );

	variationImg.Initialize( labelB.GetImageAttributes() );
	variationImgX.Initialize( labelB.GetImageAttributes() );
	variationImgY.Initialize( labelB.GetImageAttributes() );
	variationImgZ.Initialize( labelB.GetImageAttributes() );

	irtkGreyPixel* ptrA = labelA.GetPointerToVoxels();
	irtkGreyPixel* ptrB = labelB.GetPointerToVoxels();
	irtkGreyPixel* ptrMask = maskImg.GetPointerToVoxels();

	irtkRealPixel* ptrBiasX = biasImgX.GetPointerToVoxels();
	irtkRealPixel* ptrBiasY = biasImgY.GetPointerToVoxels();
	irtkRealPixel* ptrBiasZ = biasImgZ.GetPointerToVoxels();

	irtkRealPixel* ptrBias = biasImg.GetPointerToVoxels();

	irtkRealPixel* ptrVar = variationImg.GetPointerToVoxels();
	irtkRealPixel* ptrVarX = variationImgX.GetPointerToVoxels();
	irtkRealPixel* ptrVarY = variationImgY.GetPointerToVoxels();
	irtkRealPixel* ptrVarZ = variationImgZ.GetPointerToVoxels();

	double costSpatial = 0;
	double costDissimilarity = 0;
	double costSpatialDissimilarity = 0;

	double denomDissimilarity = 0;
	double denomSpatial = 0;
	double denomSpatialDissimilarity = 0;

	for( int z = ( is2D ? 0 : patchRadiusZ ) ; z < (is2D ? 1 : labelB.GetZ()-patchRadiusZ); ++z )
	{
		for( int y = patchRadiusY; y < labelB.GetY()-patchRadiusY; ++y )
		{
			for( int x = patchRadiusX; x < labelB.GetX()-patchRadiusX; ++x )
			{
				int index = getIndex(x,y,z);
				ptrBias[index] = 0;

				if( ptrB[index] == 0 && ptrA[index] == 0 ) continue;

				double minValue = -1;
				double maxSpatialVoxels = -1;
				int minX = x;
				int minY = y;
				int minZ = z;

				double mean_patchNorm_xB = 0;
				double mean_patchNorm_yB = 0;
				double mean_patchNorm_zB = 0;

				irtkImageAttributes attr = labelB.GetImageAttributes();
				attr._x = patchSizeX;
				attr._y = patchSizeY;
				attr._z = is2D ? 1 : patchSizeZ;

				irtkRealImage currentPatch( attr );
				int voxelsInCurrentPatch = 0;
				for( int dz = (is2D ? 0 : -patchRadiusZ); dz <= (is2D ? 0 : patchRadiusZ); ++dz)
				{
					for( int dy = -patchRadiusY; dy <= patchRadiusY; ++dy)
					{
						for( int dx = -patchRadiusX; dx <= patchRadiusX; ++dx)
						{
							int index_dx = getIndex(x+dx,y+dy,z+dz);
							double patchNorm_xB = 0;
							double patchNorm_yB = 0;
							double patchNorm_zB = 0;

							if( ptrB[index_dx] )
							{
								calculatePatchNormal( x+dx, y+dy, z+dz, patchNorm_xB, patchNorm_yB, patchNorm_zB );
								mean_patchNorm_xB += patchNorm_xB;
								mean_patchNorm_yB += patchNorm_yB;
								mean_patchNorm_zB += patchNorm_zB;
								voxelsInCurrentPatch++;
							}
							currentPatch.Put( dx+patchRadiusX, dy+patchRadiusY, is2D ? 0 : dz+patchRadiusZ, ptrB[index_dx] );
						}
					}
				}

				if( voxelsInCurrentPatch )
				{
					mean_patchNorm_xB /= voxelsInCurrentPatch;
					mean_patchNorm_yB /= voxelsInCurrentPatch;
					mean_patchNorm_zB /= voxelsInCurrentPatch;
				}
				minValue = comparePatches( x, y, z, x, y, z, minValue );
				minX = x;
				minY = y;
				minZ = z;
				maxSpatialVoxels = valueL1Steps[0][0];

				int L1 = 1;

				while( true && !bNoNonLocal)
				{
					bool updated = false;

					for( int s = 0; s < nrOfL1Steps[L1]; ++s )
					{
						int cx = x - xL1Steps[L1][s];
						int cy = y - yL1Steps[L1][s];
						int cz = z - zL1Steps[L1][s];
						int curSpatialVoxels = valueL1Steps[L1][s];

						if( cx >= labelB.GetX() || cy >= labelB.GetY() || cz >= labelB.GetZ() ||
									cx < 0 || cy < 0 || cz < 0 ) continue;

						int value = comparePatches( cx, cy, cz, x, y, z, minValue );

						if( minValue == value && minValue != 0 ) updated = true;

						if( value == -1 && minValue >= 0 ) continue;

						if( minValue == -1 && value == -1 )
						{
							minX = cx;
							minY = cy;
							minZ = cz;
							updated = true;
						}
						else if( minValue == -1 || value < minValue || ( value == minValue && curSpatialVoxels > maxSpatialVoxels ) )
						{
							maxSpatialVoxels = valueL1Steps[L1][s];
							minValue = value;
							minX = cx;
							minY = cy;
							minZ = cz;
							updated = true;
						}
					}
					if( !updated ) break;
					L1++;
				}

				double totVar = 0.5;
				double totVarX = 1;
				double totVarY = 1;
				double totVarZ = 1;
				if( !bNoSpatialWeighting )
				{
					if( maxNumberOfFacets == 0 )
					{
						maxNumberOfFacets = maxNumberOfEdgeFacets( currentPatch );
					}
					numberOfEdgeFacets( currentPatch, &totVar, &totVarX, &totVarY, &totVarZ );

					totVar /= maxNumberOfFacets;
					totVarX /= maxNumberOfFacets;
					totVarY /= maxNumberOfFacets;
					totVarZ /= maxNumberOfFacets;
				}
				ptrVar[getIndex(x,y,z)] = totVar;
				ptrVarX[getIndex(x,y,z)] = totVarX;
				ptrVarY[getIndex(x,y,z)] = totVarY;
				ptrVarZ[getIndex(x,y,z)] = totVarZ;

				double alpha = min(totVar, 1.0);
				double beta = 1.0 - min( totVar, 1.0);

				double tmp_costDissimilarity = getSimilarityIndex( minValue );
				double tmp_costSpatial = getSpatialIndex( minX -x, minY - y, minZ - z );
				double tmp_combined = (tmp_costDissimilarity + tmp_costSpatial ) / 2.0;

				denomDissimilarity += alpha * tmp_costDissimilarity + beta * (1.0 - tmp_costDissimilarity);
				denomSpatial += alpha * tmp_costSpatial + beta * (1.0 - tmp_costSpatial);
				denomSpatialDissimilarity += alpha * tmp_combined + beta * (1.0 - tmp_combined);

				costDissimilarity += alpha * tmp_costDissimilarity;
				costSpatial += alpha * tmp_costSpatial;
				costSpatialDissimilarity += alpha * tmp_combined;

				debugMaxL1 = max( debugMaxL1, abs(minX - x) + abs(minY - y) + abs(minZ -z) );

				ptrBiasX[getIndex(x,y,z)] = minX - x;
				ptrBiasY[getIndex(x,y,z)] = minY - y;
				ptrBiasZ[getIndex(x,y,z)] = minZ - z;

				double normBias = mean_patchNorm_xB * (minX - x) + mean_patchNorm_yB * (minY - y) + mean_patchNorm_zB * (minZ - z);

				normBias *= -1.0;

				ptrBias[getIndex(x,y,z)] = normBias;
				ptrMask[getIndex(x,y,z)] = 1;
			}
		}
	}
if( bDebug ) cerr << "Max L1: " << debugMaxL1 << endl;
if( bDebug ) cerr << "starting Bias estimation" << endl;
	double biasMean = 0;

	double biasMeanX = 0;
	double biasMeanY = 0;
	double biasMeanZ = 0;

	double biasVar = 0;

	double biasVarX = 0;
	double biasVarY = 0;
	double biasVarZ = 0;

	double biasDenom = 0;
	double biasDenomX = 0;
	double biasDenomY = 0;
	double biasDenomZ = 0;

	for( int i = 0; i < labelB.GetNumberOfVoxels(); ++i )
	{
		if( ptrMask[i] )
		{
			biasMean += ptrVar[i] * ptrBias[i];
			biasMeanX += ptrVarX[i] * ptrBiasX[i];
			biasMeanY += ptrVarY[i] * ptrBiasY[i];
			biasMeanZ += ptrVarZ[i] * ptrBiasZ[i];
			biasDenom += ptrVar[i];
			biasDenomX += ptrVarX[i];
			biasDenomY += ptrVarY[i];
			biasDenomZ += ptrVarZ[i];
		}
	}

	if( biasDenom == 0 ) biasDenom = 1;
	if( biasDenomX == 0 ) biasDenomX = 1;
	if( biasDenomY == 0 ) biasDenomY = 1;
	if( biasDenomZ == 0 ) biasDenomZ = 1;

	biasMean /= biasDenom;
	biasMeanX /= biasDenomX;
	biasMeanY /= biasDenomY;
	biasMeanZ /= biasDenomZ;

	for( int i = 0; i < labelB.GetNumberOfVoxels(); ++i )
	{
		if( ptrB[i] && ptrMask[i] )
		{
			biasVar += ptrVar[i] *(ptrBias[i] - biasMean) * (ptrBias[i] - biasMean);

			biasVarX += ptrVarX[i]* (ptrBiasX[i] - biasMeanX) * (ptrBiasX[i] - biasMeanX);
			biasVarY += ptrVarY[i] * (ptrBiasY[i] - biasMeanY) * (ptrBiasY[i] - biasMeanY);
			biasVarZ += ptrVarZ[i] * (ptrBiasZ[i] - biasMeanZ) * (ptrBiasZ[i] - biasMeanZ);
		}
	}
	biasVar /= biasDenom;
	biasVarX /= biasDenomX;
	biasVarY /= biasDenomY;
	biasVarZ /= biasDenomZ;

if( bDebug ) { cerr << "writing bias images /tmp/biasXYZ.nii.gz" << endl; biasImg.Write("/tmp/biasImageTotal.nii.gz"); biasImgX.Write("/tmp/biasImageX.nii.gz"); biasImgY.Write("/tmp/biasImageY.nii.gz"); biasImgZ.Write("/tmp/biasImageZ.nii.gz"); }

	if ( outputBias != NULL )
	{
		biasImg.Write( outputBias );
	}
	double resSpatial = costSpatial / denomSpatial;
	double resOverlap = costDissimilarity / denomDissimilarity;
	double resSpatialDissimilarity = costSpatialDissimilarity / denomSpatialDissimilarity;

	cout << resSpatialDissimilarity << " " <<  resOverlap << " " << resSpatial << " " << biasMean << " " << sqrt(biasVar) << " " << biasMeanX << " " << sqrt(biasVarX) << " " << biasMeanY << " " << sqrt(biasVarY) << " " << biasMeanZ << " " << sqrt(biasVarZ) << endl;
	return 0;
}

int getIndex( int x, int y, int z)
{
	if( x < 0 ) x = 0;
	if( y < 0 ) y = 0;
	if( z < 0 ) z = 0;
	if( x >= labelB.GetX() ) x = labelB.GetX() - 1;
	if( y >= labelB.GetY() ) y = labelB.GetY() - 1;
	if( z >= labelB.GetZ() ) z = labelB.GetZ() - 1;

	return z * labelB.GetX() * labelB.GetY() + y * labelB.GetX() + x;
}

void calculatePatchNormal( int x, int y, int z, double &xn, double &yn, double  &zn )
{
	double patchNorm_x = 0;
	double patchNorm_y = 0;
	double patchNorm_z = 0;

	double norm = 0;

	irtkRealPixel* ptrGradX = gradBX.GetPointerToVoxels();
	irtkRealPixel* ptrGradY = gradBY.GetPointerToVoxels();
	irtkRealPixel* ptrGradZ = gradBZ.GetPointerToVoxels();

	patchNorm_x = ptrGradX[getIndex(x,y,z)];
	patchNorm_y = ptrGradY[getIndex(x,y,z)];
	patchNorm_z = is2D ? 0 : ptrGradZ[getIndex(x,y,z)];

	norm = calculateVectorNorm( patchNorm_x, patchNorm_y, patchNorm_z);

	if( norm != 0 )
	{
		patchNorm_x /= norm;
		patchNorm_y /= norm;
		patchNorm_z /= norm;
	}
	else
	{
		patchNorm_x = 0;
		patchNorm_y = 0;
		patchNorm_z = 0;
	}

	xn = patchNorm_x;
	yn = patchNorm_y;
	zn = patchNorm_z;
}

void calculateDistanceTransform()
{
if( bDebug ) cerr << "Calculating distance transforms" << endl;

	irtkGreyPixel* ptrB = labelB.GetPointerToVoxels();

	// Calculating distance Map
	irtkRealImage invertB;

	invertB.Initialize( labelB.GetImageAttributes() );
	distMapB.Initialize( labelB.GetImageAttributes() );

	irtkRealPixel* ptrInvertB = invertB.GetPointerToVoxels();

	for( int i = 0; i < invertB.GetNumberOfVoxels(); ++i )
	{
		if( ptrB[i] > 0 )
		{
			ptrInvertB[i] = 0;
		}
		else
		{
			ptrInvertB[i] = 1;
		}
	}

	irtkRealImage dtInvertB;
	dtInvertB.Initialize( labelB.GetImageAttributes() );

	irtkEuclideanDistanceTransform<irtkRealPixel> *edt = new irtkEuclideanDistanceTransform<irtkRealPixel>
	(irtkEuclideanDistanceTransform<irtkRealPixel>::irtkDistanceTransform3D);

	edt->SetInput (&invertB);
	edt->SetOutput(&dtInvertB);
	edt->Run();
	delete edt;

	irtkRealPixel* ptrDtInvertB = dtInvertB.GetPointerToVoxels();
	irtkRealPixel* ptrDistMapB = distMapB.GetPointerToVoxels();
	for( int i = 0; i < labelB.GetNumberOfVoxels(); ++i )
	{
		ptrDistMapB[i] = ptrB[i] > 0 ? sqrt(ptrDtInvertB[i]) : 0;
	}

	double sigma = labelB.GetXSize();
if( bDebug ) { cerr << "blurring distance map with sigma=" << sigma << endl; }

	irtkGaussianBlurring<irtkRealPixel> filter( sigma );
	filter.SetInput(&distMapB);
	filter.SetOutput(&distMapB);
	filter.Run();

if( bDebug ) { cerr << "writing distance map: /tmp/distanceTransformB.nii.gz" << endl;  distMapB.Write("/tmp/distanceTransformB.nii.gz");}
}
void calculateGradients()
{
	calculateDistanceTransform();
if( bDebug ) cerr << "Calculating gradients" << endl;

	irtkRealImage gradBVec;
	irtkImageAttributes attr = labelB.GetImageAttributes();
	attr._t=3;
	gradBVec.Initialize(attr);

	irtkGradientImageFilter<irtkRealPixel> gradFilter(irtkGradientImageFilter<irtkRealPixel>::GRADIENT_VECTOR);
	gradFilter.SetInput( &distMapB );
	gradFilter.SetOutput( &gradBVec );
	gradFilter.SetPadding(0);
	gradFilter.Run();

	gradBX = gradBVec.GetFrame(0);
	gradBY = gradBVec.GetFrame(1);
	gradBZ = gradBVec.GetFrame(2);

if( bDebug ) { cerr << "writing gradients: /tmp/gradXYZ.nii.gz" << endl; gradBX.Write("/tmp/gradX.nii.gz"); gradBY.Write("/tmp/gradY.nii.gz"); gradBZ.Write("/tmp/gradZ.nii.gz");}
}

int maxNumberOfEdgeFacets( irtkRealImage& currentPatch )
{
	int nr = (patchSizeX - 1) * 4;
	if( !is2D ) nr *= patchSizeZ;
	if( bDebug ) cerr << "Max number of facets for " << currentPatch.GetX() << "x" << currentPatch.GetY() << "x" << currentPatch.GetZ() << " is " << nr << endl;

	return nr;
}

void numberOfEdgeFacets( irtkRealImage& currentPatch, double* totVar, double *totVarX, double *totVarY, double *totVarZ )
{
	int nr = 0;
	int nrX = 0;
	int nrY = 0;
	int nrZ = 0;
	int limX = currentPatch.GetX()-1;
	int limY = currentPatch.GetY()-1;
	int limZ = currentPatch.GetZ()-1;

	for( int x = 0; x < currentPatch.GetX(); ++x ) {
		for( int y = 0; y < currentPatch.GetY(); ++y ) {
			for( int z = 0; z < currentPatch.GetZ(); ++z ) {
				int label = currentPatch.Get(x,y,z);
				if( x < limX )
				{
					nr += label != currentPatch.Get(x+1, y, z) ? 1 : 0;
					nrX += label != currentPatch.Get(x+1, y, z) ? 1 : 0;
				}
				if( y < limY )
				{
					nr += label != currentPatch.Get(x, y+1, z) ? 1 : 0;
					nrY += label != currentPatch.Get(x, y+1, z) ? 1 : 0;
				}
				if( z < limZ && !is2D  )
				{
					nr += label != currentPatch.Get(x, y, z+1) ? 1 : 0;
					nrZ += label != currentPatch.Get(x, y, z+1) ? 1 : 0;
				}

				if( x > 0 )
				{
					nr += label != currentPatch.Get(x-1, y, z) ? 1 : 0;
					nrX += label != currentPatch.Get(x-1, y, z) ? 1 : 0;
				}
				if( y > 0 )
				{
					nr += label != currentPatch.Get(x, y-1, z) ? 1 : 0;
					nrY += label != currentPatch.Get(x, y-1, z) ? 1 : 0;
				}
				if( z > 0 && !is2D )
				{
					nr += label != currentPatch.Get(x, y, z-1) ? 1 : 0;
					nrZ += label != currentPatch.Get(x, y, z-1) ? 1 : 0;
				}
			}
		}
	}
	nr /= 2;
	nrX /= 2;
	nrY /= 2;
	nrZ /= 2;

	*totVar = nrX + nrY + nrZ;
	*totVarX = nrX;
	*totVarY = nrY;
	*totVarZ = nrZ;

	if( nr != *totVar )
	{
		cerr << nr << " Error: this should not happen! " << *totVar << endl;
	}
}

double getSimilarityIndex( int incorrectLabels )
{
	if( incorrectLabels < 0 ) return 0;

	double maxDissimilarity = patchSizeX * patchSizeY * patchSizeZ;
	if( is2D ) maxDissimilarity = patchSizeX * patchSizeY;

	if( !simTypeGaussian )
	{
		return 1.0 - incorrectLabels / maxDissimilarity;
	}
	else
	{
		double val = incorrectLabels;
		return exp( -1.0 * val  / ( 2.0 * sigmaSimilarity * sigmaSimilarity ) );
	}
}

// no need to account for voxel size here as it cancels out
double getSpatialIndex( int dx, int dy, int dz )
{
	dx = dx < 0 ? -1 * dx : dx;
	dy = dy < 0 ? -1 * dy : dy;
	dz = dz < 0 ? -1 * dz : dz;

	if( !simTypeGaussian )
	{
		int voxelsPresent = 0;
		int voxelsMax = 0;
		if( is2D )
		{
			voxelsMax = patchSizeX * patchSizeY;
			voxelsPresent = getVoxelsForDisplacement( dx, dy );
		}
		else
		{
			voxelsMax = patchSizeX * patchSizeY * patchSizeZ;
			voxelsPresent = getVoxelsForDisplacement( dx, dy, dz );
		}
		return (1.0 * voxelsPresent / voxelsMax);
	}
	else
	{
		double L1norm = is2D ? (dx + dy) : (dx + dy + dz);
		return exp( -1.0 * L1norm  / ( 2.0 * sigmaSpatial * sigmaSpatial ) );
	}
}

int getVoxelsForDisplacement( int dx, int dy, int dz )
{
	dx = abs(dx);
	dy = abs(dy);
	dz = abs(dz);

	if( is2D )
	{
		if( dx >= patchSizeX || dy >= patchSizeY ) return 0;
		return (patchSizeX - dx) * (patchSizeY - dy);
	}
	else
	{
		if( dx >= patchSizeX || dy >= patchSizeY || dz >= patchSizeZ ) return 0;
		return (patchSizeX - dx) * (patchSizeY - dy) * (patchSizeZ - dz);
	}
}

int comparePatches( int xA, int yA, int zA, int xB ,int yB, int zB, int minValue )
{
	double value = 0;

	irtkGreyPixel* ptrA = labelA.GetPointerToVoxels();
	irtkGreyPixel* ptrB = labelB.GetPointerToVoxels();

	int iA = getIndex(xA-patchRadiusX,yA-patchRadiusY,zA-patchRadiusZ);
	int iB = getIndex(xB-patchRadiusX,yB-patchRadiusY,zB-patchRadiusZ);

	if( xA-patchRadiusX < 0 || xA + patchRadiusX >= labelA.GetX() ||
			xB-patchRadiusX < 0 || xB + patchRadiusX >= labelB.GetX() ||
				yA-patchRadiusY < 0 || yA + patchRadiusY >= labelA.GetY() ||
					yB-patchRadiusY < 0 || yB + patchRadiusY >= labelB.GetY() )
		return -1;
	if( !is2D &&
			( zA-patchRadiusZ < 0 || zA + patchRadiusZ >= labelA.GetZ() ||
				zB-patchRadiusZ < 0 || zB + patchRadiusZ >= labelB.GetZ() ) )
	{
		return -1;
	}

	int incX = 1;
	int incY = labelB.GetX();
	int incZ = labelB.GetX() * labelB.GetY();

	bool isEmptyA = true;
	bool isEmptyB = true;

	for( int dz = ( is2D ? 0 : -patchRadiusZ ) ; dz <= ( is2D ? 0 : patchRadiusZ ) ; ++dz)
	{
		for( int dy = -patchRadiusY; dy <= patchRadiusY; ++dy)
		{
			for( int dx = -patchRadiusX; dx <= patchRadiusX; ++dx)
			{
				if( ptrA[iA] ) isEmptyA = false;
				if( ptrB[iB] ) isEmptyB = false;

				double isDifferentLabel = ( ptrA[iA] == ptrB[iB] ) ? 0 : 1;
				value += isDifferentLabel;
				if( value > minValue && minValue > -1 )
				{
					dx = patchRadiusX;
					dy = patchRadiusY;
					dz = is2D ? 0 : patchRadiusZ;
				}
				iA += incX;
				iB += incX;
			}
			iA -= incX * (2 * patchRadiusX+1);
			iB -= incX * (2 * patchRadiusX+1);
			iA += incY;
			iB += incY;
		}
		iA -= incY * (2 * patchRadiusY+1);
		iB -= incY * (2 * patchRadiusY+1);
		iA += incZ;
		iB += incZ;
	}
	return ( isEmptyA == isEmptyB ) ? value : -1;
}

void calculateL1Steps()
{
if( bDebug ) cerr << "Calculating L1 Steps " << endl;

	memset(nrOfL1Steps, 0, sizeof(int) * (MAXL1+1) );
	memset(xL1Steps, 0, sizeof(int) * (MAXL1+1) * 4 * (MAXL1+1)*(MAXL1+1) );
	memset(yL1Steps, 0, sizeof(int) * (MAXL1+1) * 4 * (MAXL1+1)*(MAXL1+1) );
	memset(zL1Steps, 0, sizeof(int) * (MAXL1+1) * 4 * (MAXL1+1)*(MAXL1+1) );
	memset(valueL1Steps, 0, sizeof(int) * (MAXL1+1) * 4 * (MAXL1+1)*(MAXL1+1) );

	double minVoxelSize = min( labelB.GetXSize(), min( labelB.GetYSize(), labelB.GetZSize()));

	for( int dz = -MAXL1 ; dz <= MAXL1; ++dz )
	{
		for( int dy = -MAXL1 ; dy <= MAXL1; ++dy )
		{
			for( int dx = -MAXL1 ; dx <= MAXL1; ++dx )
			{
				if( is2D && dz != 0 ) continue;

				int L1 = round( (labelB.GetXSize() * abs(dx) + labelB.GetYSize() * abs(dy) + labelB.GetZSize() * abs(dz)) / minVoxelSize);
				if( L1 > L1limit ) continue;

				xL1Steps[L1][nrOfL1Steps[L1]] = dx;
				yL1Steps[L1][nrOfL1Steps[L1]] = dy;
				zL1Steps[L1][nrOfL1Steps[L1]] = dz;
				valueL1Steps[L1][nrOfL1Steps[L1]] = getVoxelsForDisplacement(dx,dy,dz);
				nrOfL1Steps[L1]++;
			}
		}
	}

if( bDebug )
{
	for( int i = 0; i <= MAXL1; ++i )
	{
		cerr << i << ": " << nrOfL1Steps[i] << endl;
	}
}
	return;
}

double calculateVectorNorm( double x, double y, double z)
{
	double norm = x * x + y * y + z * z;
	if( norm > 0 )
	{
		return sqrt(norm);
	}
	return 0;
}

